@@ -5,25 +5,40 @@ from __future__ import annotations
 from dataclasses import dataclass, field
 from datetime import datetime, time, timedelta, timezone
 from enum import Enum, auto
+from importlib import resources
 from typing import Optional, Union
 
 from apischema import type_name
 from apischema.conversions import Conversion
-from apischema.metadata import conversion
+from apischema.metadata import conversion, fall_back_on_default
+from gql.dsl import DSLSchema
+from graphql import build_ast_schema, parse
 
 # The names in this file are from the GraphQL schema and don't always adhere to
 # the naming scheme that pylint expects.
 # pylint: disable=invalid-name
 
+SCHEMA_TEXT = resources.files(__package__).joinpath("hydrawise.graphql").read_text()
+DSL_SCHEMA = DSLSchema(build_ast_schema(parse(SCHEMA_TEXT)))
 
-def default_datetime() -> datetime:
-    """Default datetime factory for fields in this module.
 
-    Abstracted so it can be mocked out.
+def _optional_field(*args, **kwargs):
+    if "metadata" in kwargs:
+        kwargs["metadata"] |= fall_back_on_default
+    else:
+        kwargs["metadata"] = fall_back_on_default
+    return field(*args, **kwargs)
+
+
+def _now() -> datetime:
+    """Current datetime.
 
     :meta private:
     """
-    return datetime.now()
+    return datetime.now().replace(microsecond=0)
+
+
+default_datetime = _now
 
 
 def _duration_conversion(unit: str) -> conversion:
@@ -54,33 +69,35 @@ def _time_conversion() -> conversion:
         Conversion(
             lambda s: datetime.strptime(s, "%H:%M").time(), source=str, target=time
         ),
-        Conversion(lambda t: time.strftime("%H:%M"), source=time, target=str),
+        Conversion(lambda t: t.strftime("%H:%M"), source=time, target=str),
     )
 
 
 def _list_conversion(element_conversion) -> conversion:
     return conversion(
         Conversion(
-            lambda l: list(map(element_conversion.deserialization.converter, l)),
+            lambda _list: list(
+                map(element_conversion.deserialization.converter, _list)
+            ),
             source=list,
             target=list,
         ),
         Conversion(
-            lambda l: list(map(element_conversion.serialization.converter, l)),
+            lambda _list: list(map(element_conversion.serialization.converter, _list)),
             source=list,
             target=list,
         ),
     )
 
 
-class AutoEnum(Enum):
+class _AutoEnum(Enum):
     @staticmethod
     def _generate_next_value_(name, start, count, last_values):
         """Determines the value for an auto() call."""
         return name
 
 
-class StatusCodeEnum(AutoEnum):
+class StatusCodeEnum(_AutoEnum):
     """Response status codes."""
 
     OK = auto()
@@ -100,8 +117,8 @@ class StatusCodeAndSummary:
 class LocalizedValueType:
     """A localized value."""
 
-    value: float = 0.0
-    unit: str = ""
+    value: float = _optional_field(default=0.0)
+    unit: str = _optional_field(default="")
 
 
 @dataclass
@@ -109,7 +126,7 @@ class SelectedOption:
     """A generic option."""
 
     value: int = 0
-    label: Optional[str] = ""
+    label: str = _optional_field(default="")
 
 
 @dataclass
@@ -175,7 +192,7 @@ class RunTimeGroup:
     """The runtime of a watering program group."""
 
     id: int = 0
-    name: Optional[str] = ""
+    name: str = _optional_field(default="")
     duration: timedelta = field(
         metadata=_duration_conversion("minutes"), default=timedelta()
     )
@@ -185,8 +202,8 @@ class RunTimeGroup:
 class WateringPeriodicity:
     """Watering frequency description (e.g., "Every Program Start Time")."""
 
-    value: Optional[int] = 0
-    label: Optional[str] = ""
+    value: int = _optional_field(default=0)
+    label: str = _optional_field(default="")
 
 
 @dataclass
@@ -236,13 +253,13 @@ class AdvancedProgram(Program):
 
     zone_specific: bool = False
     advanced_program_id: int = 0
-    watering_frequency: Optional[ProgramWateringFrequency] = field(
+    watering_frequency: ProgramWateringFrequency = _optional_field(
         default_factory=ProgramWateringFrequency
     )
-    run_time_group: Optional[RunTimeGroup] = field(default_factory=RunTimeGroup)
+    run_time_group: RunTimeGroup = _optional_field(default_factory=RunTimeGroup)
 
 
-class AdvancedProgramDayPatternEnum(AutoEnum):
+class AdvancedProgramDayPatternEnum(_AutoEnum):
     """A value for an advanced watering program day pattern."""
 
     EVEN = auto()
@@ -269,7 +286,7 @@ class WateringSettings:
 class AdvancedWateringSettings(WateringSettings):
     """Advanced watering program settings."""
 
-    advanced_program: AdvancedProgram = None
+    advanced_program: Optional[AdvancedProgram] = None
 
 
 @dataclass
@@ -277,11 +294,11 @@ class AdvancedWateringSettings(WateringSettings):
 class TimeRange:
     """Time range units."""
 
-    valid_from: Optional[datetime] = field(
-        metadata=_timestamp_conversion(), default_factory=datetime
+    valid_from: datetime = _optional_field(
+        metadata=_timestamp_conversion(), default_factory=default_datetime
     )
-    valid_to: Optional[datetime] = field(
-        metadata=_timestamp_conversion(), default_factory=datetime
+    valid_to: datetime = _optional_field(
+        metadata=_timestamp_conversion(), default_factory=default_datetime
     )
 
 
@@ -299,14 +316,14 @@ class StandardProgramPeriodicity:
 class StandardProgram(Program):
     """A standard watering program."""
 
-    start_times: Optional[list[time]] = field(
+    start_times: list[time] = _optional_field(
         metadata=_list_conversion(_time_conversion()), default_factory=list
     )
     time_range: TimeRange = field(default_factory=TimeRange)
     ignore_rain_sensor: bool = False
     days_run: list[DaysOfWeekEnum] = field(default_factory=list)
     standard_program_day_pattern: str = ""
-    periodicity: Optional[StandardProgramPeriodicity] = field(
+    periodicity: StandardProgramPeriodicity = _optional_field(
         default_factory=StandardProgramPeriodicity
     )
 
@@ -333,15 +350,15 @@ class StandardWateringSettings(WateringSettings):
 class RunStatus:
     """Run status."""
 
-    value: int = 0
-    label: str = ""
+    value: int = _optional_field(default=0)
+    label: str = _optional_field(default="")
 
 
 @dataclass
 class ScheduledZoneRun:
     """A scheduled zone run."""
 
-    id: str = 0
+    id: str = ""
     start_time: datetime = field(
         metadata=DateTime.conversion(), default_factory=default_datetime
     )
@@ -375,7 +392,7 @@ class PastZoneRuns:
     """Previous zone runs."""
 
     last_run: Optional[ScheduledZoneRun] = None
-    runs: list[ScheduledZoneRun] = field(default_factory=list)
+    runs: list[ScheduledZoneRun] = _optional_field(default_factory=list)
 
 
 @dataclass
@@ -384,7 +401,7 @@ class ZoneStatus:
 
     relative_water_balance: int = 0
     suspended_until: Optional[datetime] = field(
-        metadata=DateTime.conversion(), default_factory=default_datetime
+        metadata=DateTime.conversion(), default=None
     )
 
 
@@ -393,10 +410,10 @@ class ZoneSuspension:
     """A zone suspension."""
 
     id: int = 0
-    start_time: datetime = field(
+    start_time: datetime = _optional_field(
         metadata=DateTime.conversion(), default_factory=default_datetime
     )
-    end_time: datetime = field(
+    end_time: datetime = _optional_field(
         metadata=DateTime.conversion(), default_factory=default_datetime
     )
 
@@ -405,21 +422,56 @@ class ZoneSuspension:
 class Zone(BaseZone):
     """A watering zone."""
 
-    watering_settings: Union[
-        AdvancedWateringSettings, StandardWateringSettings
-    ] = field(default_factory=StandardWateringSettings)
+    watering_settings: Union[AdvancedWateringSettings, StandardWateringSettings] = (
+        field(default_factory=StandardWateringSettings)
+    )
     scheduled_runs: ScheduledZoneRuns = field(default_factory=ScheduledZoneRuns)
     past_runs: PastZoneRuns = field(default_factory=PastZoneRuns)
     status: ZoneStatus = field(default_factory=ZoneStatus)
     suspensions: list[ZoneSuspension] = field(default_factory=list)
 
+    @classmethod
+    def from_json(cls, zone_json: dict) -> Zone:
+        zone = Zone(
+            id=zone_json["relay_id"],
+            number=zone_json["relay"],
+            name=zone_json["name"],
+        )
+        zone.update_with_json(zone_json)
+        return zone
+
+    def update_with_json(self, zone_json: dict) -> None:
+        current_run = None
+        next_run = None
+        suspended_until = None
+        if zone_json["time"] == 1:
+            current_run = ScheduledZoneRun(
+                remaining_time=timedelta(seconds=zone_json["run"]),
+            )
+        elif zone_json["time"] == 1576800000:
+            suspended_until = datetime.max
+        else:
+            start_time = _now() + timedelta(seconds=zone_json["time"])
+            duration = timedelta(seconds=zone_json["run"])
+            next_run = ScheduledZoneRun(
+                start_time=start_time,
+                end_time=start_time + duration,
+                normal_duration=duration,
+                duration=duration,
+            )
+        self.scheduled_runs = ScheduledZoneRuns(
+            current_run=current_run,
+            next_run=next_run,
+        )
+        self.status = ZoneStatus(suspended_until=suspended_until)
+
 
 @dataclass
 class ProgramStartTimeApplication:
     """Application of a start time to a program."""
 
     all: bool = False
-    zones: list[BaseZone] = field(default_factory=list)
+    zones: list[BaseZone] = _optional_field(default_factory=list)
 
 
 @dataclass
@@ -428,7 +480,7 @@ class ProgramStartTime:
 
     id: int = 0
     time: time = field(metadata=_time_conversion(), default_factory=time)
-    watering_days: Optional[list[AdvancedProgramDayPatternEnum]] = field(
+    watering_days: list[AdvancedProgramDayPatternEnum] = _optional_field(
         default_factory=list
     )
     application: ProgramStartTimeApplication = field(
@@ -441,7 +493,7 @@ class ControllerFirmware:
     """Information about the controller's firmware."""
 
     type: str = ""
-    version: str = ""
+    version: str = _optional_field(default="")
 
 
 @dataclass
@@ -456,11 +508,20 @@ class ControllerModel:
 class ControllerHardware:
     """Information about a controller's hardware."""
 
-    serial_number: str = ""
-    version: str = ""
-    status: str = ""
-    model: ControllerModel = field(default_factory=ControllerModel)
-    firmware: list[ControllerFirmware] = field(default_factory=list)
+    serial_number: str = _optional_field(default="")
+    version: str = _optional_field(default="")
+    status: str = _optional_field(default="")
+    model: ControllerModel = _optional_field(default_factory=ControllerModel)
+    firmware: list[ControllerFirmware] = _optional_field(default_factory=list)
+
+
+class CustomSensorTypeEnum(_AutoEnum):
+    """A value for a sensor type."""
+
+    LEVEL_OPEN = auto()
+    LEVEL_CLOSED = auto()
+    FLOW = auto()
+    THRESHOLD = auto()
 
 
 @dataclass
@@ -468,15 +529,16 @@ class SensorModel:
     """Information about a sensor model."""
 
     id: int = 0
-    name: str = ""
-    active: bool = False
-    off_level: int = 0
-    off_timer: int = 0
-    delay: timedelta = field(
+    name: str = _optional_field(default="")
+    active: bool = _optional_field(default=False)
+    off_level: int = _optional_field(default=0)
+    off_timer: int = _optional_field(default=0)
+    delay: timedelta = _optional_field(
         metadata=_duration_conversion("minutes"), default=timedelta()
     )
-    divisor: float = 0.0
-    flow_rate: float = 0.0
+    divisor: float = _optional_field(default=0.0)
+    flow_rate: float = _optional_field(default=0.0)
+    sensor_type: Optional[CustomSensorTypeEnum] = None
 
 
 @dataclass
@@ -484,14 +546,16 @@ class SensorStatus:
     """Current status of a sensor."""
 
     water_flow: Optional[LocalizedValueType] = None
-    active: Optional[bool] = False
+    active: bool = _optional_field(default=False)
 
 
 @dataclass
 class SensorFlowSummary:
     """Summary of a sensor's water flow."""
 
-    total_water_volume: LocalizedValueType = field(default_factory=LocalizedValueType)
+    total_water_volume: LocalizedValueType = _optional_field(
+        default_factory=LocalizedValueType
+    )
 
 
 @dataclass
@@ -505,21 +569,31 @@ class Sensor:
 
 
 @dataclass
-class WaterTime:
+@type_name("Sensor")
+class SensorWithFlowSummary(Sensor):
+    """A Sensor, as returned by its `flowSummary` method."""
+
+    flow_summary: Optional[SensorFlowSummary] = _optional_field(
+        default_factory=SensorFlowSummary
+    )
+
+
+@dataclass
+class _WaterTime:
     """A water time duration."""
 
-    value: timedelta = field(
+    value: timedelta = _optional_field(
         metadata=_duration_conversion("minutes"), default=timedelta()
     )
 
 
 @dataclass
-class ActualWaterTime(WaterTime):
+class ActualWaterTime(_WaterTime):
     """An actual water time duration."""
 
 
 @dataclass
-class NormalWaterTime(WaterTime):
+class NormalWaterTime(_WaterTime):
     """A normal water time duration."""
 
 
@@ -529,8 +603,12 @@ class ControllerStatus:
 
     summary: str = ""
     online: bool = False
-    actual_water_time: ActualWaterTime = field(default_factory=ActualWaterTime)
-    normal_water_time: NormalWaterTime = field(default_factory=NormalWaterTime)
+    actual_water_time: ActualWaterTime = _optional_field(
+        default_factory=ActualWaterTime
+    )
+    normal_water_time: NormalWaterTime = _optional_field(
+        default_factory=NormalWaterTime
+    )
     last_contact: Optional[DateTime] = None
 
 
@@ -553,29 +631,29 @@ class RunEvent:
 
     id: str = ""
     zone: BaseZone = field(default_factory=BaseZone)
-    standard_program: Optional[StandardProgramRef] = field(
+    standard_program: StandardProgramRef = _optional_field(
         default_factory=StandardProgramRef
     )
-    advanced_program: Optional[AdvancedProgramRef] = field(
+    advanced_program: AdvancedProgramRef = _optional_field(
         default_factory=AdvancedProgramRef
     )
     reported_start_time: Optional[datetime] = field(
-        metadata=DateTime.conversion(), default_factory=default_datetime
+        metadata=DateTime.conversion(), default=None
     )
     reported_end_time: Optional[datetime] = field(
-        metadata=DateTime.conversion(), default_factory=default_datetime
+        metadata=DateTime.conversion(), default=None
     )
-    reported_duration: Optional[int] = field(
+    reported_duration: timedelta = _optional_field(
         metadata=_duration_conversion("seconds"), default=timedelta()
     )
-    reported_status: Optional[RunStatusType] = field(default_factory=RunStatusType)
-    reported_water_usage: Optional[LocalizedValueType] = field(
+    reported_status: RunStatusType = _optional_field(default_factory=RunStatusType)
+    reported_water_usage: LocalizedValueType = _optional_field(
         default_factory=LocalizedValueType
     )
-    reported_stop_reason: Optional[RunStopReasonType] = field(
+    reported_stop_reason: RunStopReasonType = _optional_field(
         default_factory=RunStopReasonType
     )
-    reported_current: Optional[LocalizedValueType] = field(
+    reported_current: LocalizedValueType = _optional_field(
         default_factory=LocalizedValueType
     )
 
@@ -584,7 +662,18 @@ class RunEvent:
 class WateringReportEntry:
     """A Hydrawise watering report entry."""
 
-    run_event: Optional[RunEvent] = field(default_factory=RunEvent)
+    run_event: RunEvent = _optional_field(
+        default_factory=RunEvent, metadata=fall_back_on_default
+    )
+
+
+@dataclass
+class MasterValve:
+    """A master valve setting for a controller."""
+
+    zone_number: Optional[SelectedOption] = None
+    delay: Optional[int] = None
+    post_timer: Optional[int] = None
 
 
 @dataclass
@@ -592,20 +681,46 @@ class Controller:
     """A Hydrawise controller."""
 
     id: int = 0
-    name: str = ""
-    software_version: str = ""
+    name: str = _optional_field(default="")
+    software_version: str = _optional_field(default="")
     hardware: ControllerHardware = field(default_factory=ControllerHardware)
-    last_contact_time: datetime = field(
+    last_contact_time: datetime = _optional_field(
         metadata=DateTime.conversion(), default_factory=default_datetime
     )
-    last_action: datetime = field(
+    last_action: datetime = _optional_field(
         metadata=DateTime.conversion(), default_factory=default_datetime
     )
-    online: bool = False
-    sensors: list[Sensor] = field(default_factory=list)
-    zones: list[Zone] = field(default_factory=list)
-    permitted_program_start_times: list[ProgramStartTime] = field(default_factory=list)
-    status: ControllerStatus = None
+    online: bool = _optional_field(default=False)
+    sensors: list[Sensor] = _optional_field(default_factory=list)
+    zones: list[Zone] = _optional_field(default_factory=list)
+    master_zone: Optional[MasterValve] = None
+    permitted_program_start_times: list[ProgramStartTime] = _optional_field(
+        default_factory=list
+    )
+    status: Optional[ControllerStatus] = None
+
+    @classmethod
+    def from_json(cls, controller_json: dict) -> Controller:
+        controller = Controller(
+            id=controller_json["controller_id"],
+            name=controller_json["name"],
+            hardware=ControllerHardware(
+                serial_number=controller_json["serial_number"],
+            ),
+        )
+        controller.update_with_json(controller_json)
+        return controller
+
+    def update_with_json(self, controller_json: dict) -> None:
+        self.last_contact_time = datetime.fromtimestamp(controller_json["last_contact"])
+        self.online = True
+
+
+@dataclass
+class UnitsSummary:
+    """Summary of user unit preferences."""
+
+    units_name: str = ""
 
 
 @dataclass
@@ -615,11 +730,12 @@ class User:
     id: int = 0
     customer_id: int = 0
     name: str = ""
-    email: str = ""
-    controllers: list[Controller] = field(default_factory=list)
+    email: str = _optional_field(default="")
+    controllers: list[Controller] = _optional_field(default_factory=list)
+    units: UnitsSummary = field(default_factory=UnitsSummary)
 
 
-class DaysOfWeekEnum(AutoEnum):
+class DaysOfWeekEnum(_AutoEnum):
     """All days of the week."""
 
     SUNDAY = auto()
@@ -629,3 +745,27 @@ class DaysOfWeekEnum(AutoEnum):
     THURSDAY = auto()
     FRIDAY = auto()
     SATURDAY = auto()
+
+
+@dataclass
+class ControllerWaterUseSummary:
+    """Water use summary for a controller.
+
+    Active use means water use during a scheduled or manual zone run.
+    Inactive use means water use when no zone was actively running. This can happen when
+    faucets (i.e., garden hoses) are installed downstream of the flow meter. Water use
+    is only reported in the presence of a flow sensor. Active watering time is always
+    reported.
+    """
+
+    _pydrawise_type = True
+
+    total_active_time: timedelta = field(
+        metadata=_duration_conversion("seconds"), default=timedelta()
+    )
+    active_time_by_zone_id: dict[int, timedelta] = field(default_factory=dict)
+    total_use: Optional[float] = None
+    total_active_use: Optional[float] = None
+    total_inactive_use: Optional[float] = None
+    active_use_by_zone_id: dict[int, float] = field(default_factory=dict)
+    unit: Optional[str] = None
