@@ -4,13 +4,26 @@ from __future__ import annotations
 
 from collections import namedtuple
 from dataclasses import fields, is_dataclass
-from typing import Iterator, List, Type, Union, get_args, get_origin, get_type_hints
+from typing import (
+    TYPE_CHECKING,
+    Iterator,
+    List,
+    Union,
+    get_args,
+    get_origin,
+    get_type_hints,
+)
 
 from apischema import deserialize as _deserialize
 from apischema.metadata.keys import CONVERSION_METADATA, SKIP_METADATA
 from apischema.type_names import get_type_name
 from apischema.utils import to_camel_case
-from gql.dsl import DSLField, DSLInlineFragment, DSLSchema
+from gql.dsl import DSLField, DSLInlineFragment
+
+from .schema import DSL_SCHEMA
+
+if TYPE_CHECKING:
+    from _typeshed import DataclassInstance
 
 # For compatibility with < python 3.10.
 NoneType = type(None)
@@ -28,7 +41,9 @@ def deserialize(*args, **kwargs):
 _Field = namedtuple("_Field", ["name", "types"])
 
 
-def _fields(cls, skip: list[str]) -> Iterator[_Field]:
+def _fields(
+    cls: DataclassInstance | type[DataclassInstance], skip: list[str]
+) -> Iterator[_Field]:
     """Returns _Field objects for every field on the given dataclass.
 
     :meta private:
@@ -53,12 +68,18 @@ def _fields(cls, skip: list[str]) -> Iterator[_Field]:
         if origin == Union:
             # Drop None from Optional fields.
             field_types = set(get_args(field_type)) - {NoneType}
-            if len(field_types) == 1:
-                [field_type] = field_types
-            else:
+
+            # Actual unions just yield the union.
+            if len(field_types) > 1:
                 yield _Field(f.name, list(field_types))
                 continue
-        elif origin in (List, list):
+
+            # If we have only one type left after dropping None, this could
+            # still be a list. Perform the normal extraction routine.
+            [field_type] = field_types
+            origin = get_origin(field_type)
+
+        if origin in (List, list):
             # Extract the contained type.
             # We assume all list types are uniform.
             [field_type] = get_args(field_type)
@@ -67,7 +88,8 @@ def _fields(cls, skip: list[str]) -> Iterator[_Field]:
 
 
 def get_selectors(
-    ds: DSLSchema, cls: Type, skip_fields: list[str] | None = None
+    cls: DataclassInstance | type[DataclassInstance],
+    skip_fields: list[str] | None = None,
 ) -> list[DSLField]:
     """Constructs GraphQL selectors for the given dataclass.
 
@@ -76,14 +98,12 @@ def get_selectors(
     ret = []
     skip_now, skip_later = parse_skip(skip_fields or [])
     for f in _fields(cls, skip_now):
-        dsl_field = getattr(getattr(ds, get_type_name(cls).graphql), f.name)
+        dsl_field = getattr(getattr(DSL_SCHEMA, get_type_name(cls).graphql), f.name)  # type: ignore[arg-type]
         if len(f.types) == 1:
             [f_type] = f.types
             if is_dataclass(f_type):
                 f_skip = skip_later.get(f.name, [])
-                ret.append(
-                    getattr(dsl_field, "select")(*get_selectors(ds, f_type, f_skip))
-                )
+                ret.append(getattr(dsl_field, "select")(*get_selectors(f_type, f_skip)))
             else:
                 ret.append(dsl_field)
         else:
@@ -94,8 +114,8 @@ def get_selectors(
                     raise NotImplementedError
                 sel_args.append(
                     DSLInlineFragment()
-                    .on(getattr(ds, get_type_name(f_type).graphql))
-                    .select(*get_selectors(ds, f_type))
+                    .on(getattr(DSL_SCHEMA, get_type_name(f_type).graphql))  # type: ignore[arg-type]
+                    .select(*get_selectors(f_type))
                 )
             ret.append(getattr(dsl_field, "select")(*sel_args))
     return ret
@@ -109,8 +129,9 @@ def parse_skip(skip: list[str] | None = None) -> tuple[list[str], dict[str, list
 
     :meta private:
     """
-    now, later = [], {}
-    for item in skip:
+    now: list[str] = []
+    later: dict[str, list[str]] = {}
+    for item in skip or []:
         field, _, descendants = item.partition(".")
         if descendants:
             later.setdefault(field, []).append(descendants)
